from functools import partial
import os
from pathlib import Path
from typing import List, Tuple, Any, Dict, Optional
import torch
import uuid
import hashlib
from hf_ehr.config import PATH_TO_TOKENIZER_CLMBR_v8_CONFIG

def convert_lab_value_to_token_from_ranges(code: str, unit: str, value: float, ranges: List[Tuple[float, float]], is_tokenize_out_of_range: bool = False) -> Optional[str]:
    # Given a list of ranges (i.e. tuples of [start, end] values), remaps the code to the index in the `ranges` array corresponds
    # to this code's value, i.e. "code" => "{code} || {idx}"
    # If the value doesn't fit in any of the ranges, returns the code itself, i.e. "{code}"
    for idx, (start_val, end_val) in enumerate(ranges):
        if start_val <= value <= end_val:
            return f"{code} || {unit} || R{idx + 1}" # "STANFORD_OBS/123 | mmol | R3"
    # Token is out of range, so decide whether to return `None` or special `R0` code
    if is_tokenize_out_of_range:
        return f"{code} || {unit} || R0" # "STANFORD_OBS/123 | mmol | R0"
    else:
        return None

def convert_lab_value_to_token_from_quantiles(code: str, unit: str, value: float, quantiles: List[float], is_tokenize_out_of_range: bool = False) -> Optional[str]:
    # Note: If we have Q1, Q2, Q3, Q4, then `len(quantiles) == 3` b/c have [0.25, 0.5, 0.75]
    for q_idx, q in enumerate(quantiles):
        if value <= q: 
            return get_lab_value_token_name(code, unit, str(q_idx + 1))
    # Token is out of range, so decide whether to return `None` or special `R0` code
    if is_tokenize_out_of_range:
        return get_lab_value_token_name(code, unit, "0") # "STANFORD_OBS/123 | mmol | Q0"
    else:
        return None

def get_lab_value_token_name(code: str, unit: str, quantile: str) -> str:
    return f"{code} || {unit} || Q{quantile}" # "STANFORD_OBS/123 | mmol | Q4"

def _get_linear_schedule_with_warmup_lr_lambda(current_step: int, 
                                               *, 
                                               num_warmup_steps: int, 
                                               num_decay_steps: int,
                                               initial_lr: float,
                                               peak_lr: float,
                                               final_lr: float) -> float:
    """Note that this needs to return a multiplier on `initial_lr` as set in the optimizer"""
    if current_step < num_warmup_steps:
        # Linear warmup from `initial_lr` to `peak_lr`
        new_lr: float = (peak_lr - initial_lr) / num_warmup_steps * current_step + initial_lr
    elif current_step < num_warmup_steps + num_decay_steps:
        # Linear decay from `peak_lr` to `final_lr`
        new_lr: float = (final_lr - peak_lr) / num_decay_steps * (current_step - num_warmup_steps) + peak_lr
    else:
        # Plateau at `final_lr`
        new_lr: float = final_lr
    multiplier: float = new_lr / peak_lr
    return multiplier

def lr_warmup_with_constant_plateau(optimizer, 
                                    num_warmup_steps: int, 
                                    num_decay_steps: int,
                                    initial_lr: float,
                                    final_lr: float, 
                                    last_epoch: int = -1):
    """
    Create a schedule with a learning rate that decreases linearly from the peak lr set in the optimizer to `final_lr`, after
    a warmup period during which it increases linearly from `initial_lr` to the peak lr set in the optimizer.

    Args:
        optimizer ([`~torch.optim.Optimizer`]):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.
        num_decay_steps (`int`):
            The total number of steps to decay the lr.
        initial_lr (`float`):
            The initial learning rate before the warmup phase
        final_lr (`float`):
            The final learning rate after the warmup and decay phases that we plateau at
        last_epoch (`int`, *optional*, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """
    assert num_warmup_steps > 0, f"num_warmup_steps must be > 0, got {num_warmup_steps}"
    assert num_decay_steps > 0, f"num_decay_steps must be > 0, got {num_decay_steps}"

    peak_lr: float = optimizer.param_groups[0]['lr']
    
    lr_lambda = partial(
        _get_linear_schedule_with_warmup_lr_lambda,
        num_warmup_steps=num_warmup_steps,
        num_decay_steps=num_decay_steps,
        initial_lr=initial_lr,
        peak_lr=peak_lr,
        final_lr=final_lr,
    )
    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda, last_epoch)

def hash_string_to_uuid(input: Any) -> str:
    """Create a MD5 hash of the stringified input"""
    input_string: str = str(input)
    md5_hash = hashlib.md5(input_string.encode()).hexdigest()
    
    # Generate a UUID from the MD5 hash
    generated_uuid = uuid.UUID(md5_hash)
    
    return str(generated_uuid)

def load_config_from_ckpt(ckpt) -> Dict[str, Any]:
    """Given a model checkpoint (from torch.load()), returns its config"""
    config = ckpt['hyper_parameters']['config']
    def recurse(d: Dict[str, Any]):
        for k, v in d.items():
            if v == 'None':
                d[k] = None
            elif isinstance(v, dict):
                recurse(v)
    recurse(config)
    return config

def load_tokenizer_old_from_config(config):
    """Load tokenizer from config."""
    from hf_ehr.data.tokenization_old import FEMRTokenizer, DescTokenizer
    from transformers import AutoTokenizer
    
    # Load config
    tokenizer__excluded_vocabs: Optional[List[str]] = config.data.tokenizer.excluded_vocabs
    tokenizer__min_code_count: Optional[int] = getattr(config.data.tokenizer, 'min_code_count', None)
    tokenizer__is_remap_numerical_codes: bool = getattr(config.data.tokenizer, 'is_remap_numerical_codes', False)
    tokenizer__is_clmbr: bool = getattr(config.data.tokenizer, 'is_clmbr', False)
    tokenizer__is_remap_codes_to_desc: bool = getattr(config.data.tokenizer, 'is_remap_codes_to_desc', False)
    tokenizer__desc_emb_tokenizer: bool = getattr(config.data.tokenizer, 'desc_emb_tokenizer', False)
    tokenizer__path_to_code_2_detail: str = config.data.tokenizer.path_to_code_2_detail.replace('/local-scratch/nigam/users/hf_ehr/', '/share/pi/nigam/mwornow/hf_ehr/cache/')

    if tokenizer__is_clmbr:
        # CLMBR
        tokenizer = FEMRTokenizer(tokenizer__path_to_code_2_detail, 
                                  excluded_vocabs=tokenizer__excluded_vocabs,
                                  is_remap_numerical_codes=tokenizer__is_remap_numerical_codes,
                                  min_code_count=tokenizer__min_code_count)
    elif tokenizer__is_remap_codes_to_desc:
        # DescTokenizer
        tokenizer = DescTokenizer(AutoTokenizer.from_pretrained(tokenizer__desc_emb_tokenizer))
    else:
        # FEMRTokenizer
        tokenizer = FEMRTokenizer(tokenizer__path_to_code_2_detail, 
                                    excluded_vocabs=tokenizer__excluded_vocabs,
                                    is_remap_numerical_codes=tokenizer__is_remap_numerical_codes,
                                    min_code_count=tokenizer__min_code_count)
    return tokenizer

def load_tokenizer_from_config(config):
    """Load tokenizer from config."""
    from hf_ehr.data.tokenization import CookbookTokenizer, CLMBRTokenizer, DescTokenizer, CEHRTokenizer
    from transformers import AutoTokenizer
    
    # Load config
    name = config.data.tokenizer.name
    path_to_config: str = config.data.tokenizer.path_to_config
    tokenizer__excluded_vocabs: Optional[List[str]] = getattr(config.data.tokenizer, 'excluded_vocabs', [])
    tokenizer__min_code_count: Optional[int] = getattr(config.data.tokenizer, 'min_code_count', None)
    tokenizer__is_remap_numerical_codes: bool = getattr(config.data.tokenizer, 'is_remap_numerical_codes', False)
    tokenizer__desc_emb_tokenizer: bool = getattr(config.data.tokenizer, 'desc_emb_tokenizer', False)

    if name == 'CLMBRTokenizer':
        # CLMBR
        tokenizer = CLMBRTokenizer(path_to_config)
    elif name == 'DescTokenizer':
        # DescTokenizer
        tokenizer = DescTokenizer(path_to_config, AutoTokenizer.from_pretrained(tokenizer__desc_emb_tokenizer))
    elif name == 'CookbookTokenizer':
        # CookbookTokenizer
        tokenizer = CookbookTokenizer(path_to_config, 
                                        excluded_vocabs=tokenizer__excluded_vocabs,
                                        is_remap_numerical_codes=tokenizer__is_remap_numerical_codes,
                                        min_code_count=tokenizer__min_code_count)
    elif name == 'CEHRTokenizer':
        # CEHRTokenizer
        tokenizer = CEHRTokenizer(path_to_config, config.get('data', {}).get('tokenizer', {}).get('metadata', {}))
    return tokenizer

def load_tokenizer_old_from_path(path_to_ckpt: str):
    """Given a path to a model checkpoint, load the tokenizer."""
    ckpt: Dict[str, Any] = torch.load(path_to_ckpt, map_location='cpu', weights_only=False)
    config: Dict[str, Any] = load_config_from_ckpt(ckpt)
    return load_tokenizer_old_from_config(config)

def load_tokenizer_from_path(path_to_ckpt: str):
    """Given a path to a model checkpoint, load the tokenizer."""
    ckpt: Dict[str, Any] = torch.load(path_to_ckpt, map_location='cpu', weights_only=False)
    config: Dict[str, Any] = load_config_from_ckpt(ckpt)
    return load_tokenizer_from_config(config)

def load_config_from_path(path_to_ckpt: str) -> Dict[str, Any]:
    """Given a path to a model checkpoint, load the config."""
    ckpt: Dict[str, Any] = torch.load(path_to_ckpt, map_location='cpu', weights_only=False)
    config: Dict[str, Any] = load_config_from_ckpt(ckpt)
    return config


def load_model_old_from_path(path_to_ckpt: str) -> torch.nn.Module:
    """Given a path to a model checkpoint, load the model."""
    from hf_ehr.models.gpt import GPTLanguageModel
    from hf_ehr.models.bert import BERTLanguageModel
    from hf_ehr.models.hyena import HyenaLanguageModel
    from hf_ehr.models.mamba import MambaLanguageModel
    from hf_ehr.models.llama import LlamaLanguageModel
    from hf_ehr.models.t5 import T5LanguageModel

    # Load checkpoint
    ckpt: Dict[str, Any] = torch.load(path_to_ckpt, map_location='cpu', weights_only=False)
    config: Dict[str, Any] = load_config_from_ckpt(ckpt)
    
    # Load tokenizer
    tokenizer = load_tokenizer_old_from_path(path_to_ckpt)

    # Determine type of model based on config.model.name
    model_map = {
        'bert': BERTLanguageModel,
        'gpt': GPTLanguageModel,
        'hyena': HyenaLanguageModel,
        'mamba': MambaLanguageModel,
        'llama': LlamaLanguageModel,
        't5': T5LanguageModel
    }
    model_name: str = config['model']['name']
    model_class = next((m for k, m in model_map.items() if k in model_name), None)
    if not model_class: raise ValueError(f"Model `{model_name}` not supported.")

    # Load model
    model = model_class(**ckpt['hyper_parameters'], vocab_size=tokenizer.vocab_size, pad_token_id=tokenizer.pad_token_id)
    model.load_state_dict(ckpt['state_dict'])
    return model

def load_ckpt(path_to_ckpt: str) -> Dict[str, Any]:
    """Given a path to a model checkpoint, load the checkpoint."""
    ckpt = torch.load(path_to_ckpt, map_location='cpu', weights_only=False)
    return ckpt

def load_model_from_path(path_to_ckpt: str) -> torch.nn.Module:
    """Given a path to a model checkpoint, load the model."""
    from hf_ehr.models.gpt import GPTLanguageModel
    from hf_ehr.models.bert import BERTLanguageModel
    from hf_ehr.models.hyena import HyenaLanguageModel
    from hf_ehr.models.mamba import MambaLanguageModel
    from hf_ehr.models.llama import LlamaLanguageModel
    from hf_ehr.models.t5 import T5LanguageModel

    # Load checkpoint
    ckpt: Dict[str, Any] = load_ckpt(path_to_ckpt)
    config = load_config_from_ckpt(ckpt)
    
    # Load tokenizer
    tokenizer = load_tokenizer_from_path(path_to_ckpt)

    # Determine type of model based on config.model.name
    model_map = {
        'bert': BERTLanguageModel,
        'gpt': GPTLanguageModel,
        'hyena': HyenaLanguageModel,
        'mamba': MambaLanguageModel,
        'llama': LlamaLanguageModel,
        't5': T5LanguageModel
    }
    model_name: str = config['model']['name']
    model_class = next((m for k, m in model_map.items() if k in model_name), None)
    if not model_class: raise ValueError(f"Model `{model_name}` not supported.")

    # Load model
    model = model_class(**ckpt['hyper_parameters'], vocab_size=tokenizer.vocab_size, pad_token_id=tokenizer.pad_token_id)
    model.load_state_dict(ckpt['state_dict'])
    return model

def get_most_recent_ckpt_from_output_dir(path_to_output_dir: str) -> Optional[str]:
    """NOTE: Not used currently, but could be a useful helper function"""
    path_to_ckpts = os.path.join(path_to_output_dir, 'ckpts')
    if os.path.exists(path_to_ckpts):
        # Loop through all ckpt files, choose most recent one
        ckpt_files: List = list(Path(path_to_ckpts).glob('*.ckpt'))
        if ckpt_files is not None and len(ckpt_files) > 0:
            max_ckpt, max_tokens = None, None
            for ckpt_file in ckpt_files:
                path_to_ckpt = os.path.join(path_to_ckpts, ckpt_file)
                ckpt = torch.load(path_to_ckpt, map_location='cpu', weights_only=False)
                if max_tokens is None or ckpt['train_total_tokens_nonPAD'] > max_tokens:
                    max_ckpt = path_to_ckpt
                    max_tokens = ckpt['train_total_tokens_nonPAD']
            assert max_ckpt is not None, f"Error -- max_ckpt is None. Couldn't find most recent ckpt."
            return max_ckpt
    return None


import datetime
import json
import socket
import os
from typing import TypedDict, Dict, Optional, List, Any, Literal, Union, Tuple, Callable
from omegaconf import DictConfig, OmegaConf
from loguru import logger
from dataclasses import dataclass, asdict, field
import logging
from tqdm import tqdm

SPLIT_SEED: int = 97
SPLIT_TRAIN_CUTOFF: float = 70
SPLIT_VAL_CUTOFF: float = 85

H100_BASE_DIR: str = '/local-scratch/nigam/users/hf_ehr/'
A100_BASE_DIR: str = '/local-scratch/nigam/hf_ehr/'
V100_BASE_DIR: str = '/local-scratch/nigam/hf_ehr/'
GPU_BASE_DIR: str = '/share/pi/nigam/data/'
SHAHLAB_SECURE_BASE_DIR: str = '/home/migufuen/hf_ehr/data/'

PATH_TO_CACHE_DIR: str = '/share/pi/nigam/mwornow/hf_ehr/cache/'
PATH_TO_RUNS_DIR: str = os.path.join(PATH_TO_CACHE_DIR, 'runs/')

# Datasets
PATH_TO_DATASET_CACHE_DIR = os.path.join(PATH_TO_CACHE_DIR, 'dataset/')
PATH_TO_FEMR_EXTRACT_v9 = '/share/pi/nigam/data/som-rit-phi-starr-prod.starr_omop_cdm5_deid_2023_08_13_extract_v9'
PATH_TO_FEMR_EXTRACT_v8 = '/share/pi/nigam/data/som-rit-phi-starr-prod.starr_omop_cdm5_deid_2023_02_08_extract_v8_no_notes'
PATH_TO_FEMR_EXTRACT_MIMIC4 = '/share/pi/nigam/data/femr_mimic_4_extract'
PATH_TO_MEDS_EXTRACT_DEV = '/share/pi/nigam/mwornow/meds_dev'

# Tokenizers
PATH_TO_TOKENIZERS_DIR: str = os.path.join(PATH_TO_CACHE_DIR, 'tokenizers/')
PATH_TO_TOKENIZER_COOKBOOK_v8_DIR: str = os.path.join(PATH_TO_TOKENIZERS_DIR, 'cookbook_v8/')
PATH_TO_TOKENIZER_COOKBOOK_DEBUG_v8_DIR: str = os.path.join(PATH_TO_TOKENIZERS_DIR, 'cookbook_debug/')
PATH_TO_TOKENIZER_COOKBOOK_MIMIC4_DIR: str = os.path.join(PATH_TO_TOKENIZERS_DIR, 'cookbook_mimic4/')
PATH_TO_TOKENIZER_CLMBR_v8_DIR: str = os.path.join(PATH_TO_TOKENIZERS_DIR, 'clmbr_v8/')
PATH_TO_TOKENIZER_CEHR_v8_DIR: str = os.path.join(PATH_TO_TOKENIZERS_DIR, 'cehr_v8/')
PATH_TO_TOKENIZER_DESC_v8_DIR: str = os.path.join(PATH_TO_TOKENIZERS_DIR, 'desc_v8/')
PATH_TO_TOKENIZER_COOKBOOK_v8_CONFIG: str = os.path.join(PATH_TO_TOKENIZER_COOKBOOK_v8_DIR, 'tokenizer_config.json')
PATH_TO_TOKENIZER_COOKBOOK_DEBUG_v8_CONFIG: str = os.path.join(PATH_TO_TOKENIZER_COOKBOOK_DEBUG_v8_DIR, 'tokenizer_config.json')
PATH_TO_TOKENIZER_COOKBOOK_MIMIC4_CONFIG: str = os.path.join(PATH_TO_TOKENIZER_COOKBOOK_MIMIC4_DIR, 'tokenizer_config.json')
PATH_TO_TOKENIZER_CLMBR_v8_CONFIG: str = os.path.join(PATH_TO_TOKENIZER_CLMBR_v8_DIR, 'tokenizer_config.json')
PATH_TO_TOKENIZER_CEHR_v8_CONFIG: str = os.path.join(PATH_TO_TOKENIZER_CEHR_v8_DIR, 'tokenizer_config.json')
PATH_TO_TOKENIZER_DESC_v8_CONFIG: str = os.path.join(PATH_TO_TOKENIZER_DESC_v8_DIR, 'tokenizer_config.json')

def wrapper_with_logging(func: Callable, func_name: str, *args: Any, **kwargs: Any) -> None:
    """
    Wrapper function to log the execution of another function and optionally handle multiprocessing.
    
    Parameters:
    - func: The function to be executed.
    - func_name: The name of the function, used for logging.
    - *args, **kwargs: Arguments and keyword arguments for the function.
    
    Returns:
    - None
    """
    logging.info(f"Starting {func_name} with args={args} and kwargs={kwargs}...")
    
    try:
        # Call the function directly, let the function itself manage multiprocessing if needed
        func(*args, **kwargs)
        logging.info(f"Finished {func_name} successfully.")
    except Exception as e:
        logging.error(f"Error in {func_name}: {str(e)}")
        raise

@dataclass()
class Event():
    code: str # LOINC/1234
    value: Optional[Any] = None # 123.45 or 'YES' or None
    unit: Optional[str] = None # mg/dL or None
    start: Optional[datetime.datetime] = None # 2023-08-13
    end: Optional[datetime.datetime] = None # 2023-08-13
    omop_table: Optional[str] = None  # 'measurement' or 'observation' or None
    
    def to_dict(self):
        return asdict(self)

#############################################
#
# Token Stats
#
#############################################
@dataclass()
class TCEStat():
    type: Literal['count_occurrences', 'count_patients', 'ppl'] # type of this stat

    def to_dict(self) -> dict:
        return asdict(self)
    
@dataclass()
class CountOccurrencesTCEStat(TCEStat):
    # Counts total # of occurrences of token in dataset split
    dataset: Optional[str] = None
    split: Optional[str] = None
    count: Optional[str] = None
    type: str = 'count_occurrences'

@dataclass()
class CountPatientsTCEStat(TCEStat):
    # Counts total # of unique patients with this token in dataset split
    dataset: Optional[str] = None
    split: Optional[str] = None
    count: Optional[int] = None
    type: str = 'count_patients'

@dataclass()
class PPLTCEStat(TCEStat):
    # Record the average perplexity of the token in the dataset split
    dataset: Optional[str] = None
    split: Optional[str] = None
    model: Optional[str] = None
    ppl: Optional[float] = None
    type: str = 'ppl'

#############################################
#
# Token Types
#
#############################################
@dataclass()
class TokenizerConfigEntry():
    code: str # LOINC/1234 -- raw code
    type: Literal['numerical_range', 'categorical', 'code'] # type of this token
    description: Optional[str] = None # 'Glucose' -- description of the code
    tokenization: Dict[str, Any] = field(default_factory=dict) # various info helpful for tokenizing
    stats: List[TCEStat] = field(default_factory=list) # various stats about this token

    def to_dict(self) -> dict:
        return asdict(self)

    def to_token(self):
        raise NotImplementedError("Subclasses must implement this method.")
    
    def get_stat(self, type_: str, settings: Optional[dict])-> Optional[TCEStat]:
        """Returns stat with specified type (required) and settings (optional)"""
        for s in self.stats:
            if s.type == type_:
                if settings is not None:
                    for key, val in settings.items():
                        if getattr(s, key) != val:
                            break
                return s
        return None

@dataclass()
class CodeTCE(TokenizerConfigEntry):
    type: str = 'code'

    def to_token(self) -> str:
        # LOINC/1234
        return f"{self.code}"

@dataclass()
class NumericalRangeTCE(TokenizerConfigEntry):
    type: str = 'numerical_range'
    tokenization: Dict[str, Any] = field(default_factory=lambda: {
        'unit' : None, 'range_start': None, 'range_end' : None, 
    })
    
    def to_token(self)-> str:
        # LOINC/1234 || mg/dL || 0.0 - 100.0
        return f"{self.code} || {self.tokenization['unit']} || {self.tokenization['range_start']} - {self.tokenization['range_end']}"

@dataclass()
class CategoricalTCE(TokenizerConfigEntry):
    type: str = 'categorical'
    tokenization: Dict[str, Any] = field(default_factory=lambda: {
        'categories': [] 
    })

    def to_token(self)-> str:
        # LOINC/1234 || category1,category2,category3
        return f"{self.code} || {','.join(self.tokenization['categories'])}"


#############################################
#
# Tokenizer config helpers
#
#############################################
def save_tokenizer_config_to_path(path_to_tokenizer_config: str, tokenizer_config: List[TokenizerConfigEntry], metadata: Optional[Dict] = None) -> None:
    """Given a path to a `tokenizer_config.json` file, saves the JSON config to disk."""
    json.dump({
        'timestamp' : str(datetime.datetime.now().isoformat()),
        'metadata' : metadata if metadata else {},
        'tokens' : [ x.to_dict() for x in tokenizer_config ], # NOTE: Takes ~30 seconds for 1.5M tokens
    }, open(path_to_tokenizer_config, 'w'), indent=2) # NOTE: Saving takes a few minutes

def load_tokenizer_config_and_metadata_from_path(path_to_tokenizer_config: str) -> Tuple[List[TokenizerConfigEntry], Dict[str, Any]]:
    return load_tokenizer_config_from_path(path_to_tokenizer_config, is_return_metadata=True) # type: ignore

def load_tokenizer_config_from_path(path_to_tokenizer_config: str, is_return_metadata: bool = False) -> Union[List[TokenizerConfigEntry], Tuple[List[TokenizerConfigEntry], Dict[str, Any]]]:
    """Given a path to a `tokenizer_config.json` file, loads the JSON config and parses into Python objects."""
    raw_data = json.load(open(path_to_tokenizer_config, 'r'))
    raw_metadata: Dict[str, Any] = raw_data['metadata']
    raw_tokens: Dict[str, Any] = raw_data['tokens']
    
    # Parse token Dict => TokenizerConfigEntry objects
    config: List[TokenizerConfigEntry] = []
    for entry in raw_tokens:
        raw_stats = entry.pop('stats')
        stats: List[TCEStat] = []
        for stat in raw_stats:
            if stat['type'] == 'count_occurrences':
                stats.append(CountOccurrencesTCEStat(**stat))
            elif stat['type'] == 'count_patients':
                stats.append(CountPatientsTCEStat(**stat))
            elif stat['type'] == 'ppl':
                stats.append(PPLTCEStat(**stat))
            else:
                raise ValueError(f"Unknown stat type: {stat}")
        if entry['type'] == 'code':
            config.append(CodeTCE(**entry, stats=stats))
        elif entry['type'] == 'numerical_range':
            config.append(NumericalRangeTCE(**entry, stats=stats))
        elif entry['type'] == 'categorical':
            config.append(CategoricalTCE(**entry, stats=stats))
        else:
            raise ValueError(f"Unknown token type: {entry}")
    if is_return_metadata:
        return config, raw_metadata
    else:
        return config

#############################################
#
# Evaluations
#
#############################################

EHRSHOT_LABELING_FUNCTION_2_PAPER_NAME = {
    # Guo et al. 2023
    "guo_los": "Long LOS",
    "guo_readmission": "30-day Readmission",
    "guo_icu": "ICU Admission",
    # New diagnosis
    "new_pancan": "Pancreatic Cancer",
    "new_celiac": "Celiac",
    "new_lupus": "Lupus",
    "new_acutemi": "Acute MI",
    "new_hypertension": "Hypertension",
    "new_hyperlipidemia": "Hyperlipidemia",
    # Instant lab values
    "lab_thrombocytopenia": "Thrombocytopenia",
    "lab_hyperkalemia": "Hyperkalemia",
    "lab_hypoglycemia": "Hypoglycemia",
    "lab_hyponatremia": "Hyponatremia",
    "lab_anemia": "Anemia",
    # # Custom tasks
    "chexpert": "Chest X-ray Findings",
    # MIMIC-IV tasks
    "mimic4_los" : "Long LOS (MIMIC-IV)",
    "mimic4_readmission" : "30-day Readmission (MIMIC-IV)",
    "mimic4_mortality" : "Inpatient Mortality (MIMIC-IV)",
}

EHRSHOT_TASK_GROUP_2_PAPER_NAME = {
    "operational_outcomes": "Operational Outcomes",
    "lab_values": "Anticipating Lab Test Results",
    "new_diagnoses": "Assignment of New Diagnoses",
    "chexpert": "Anticipating Chest X-ray Findings",
}

EHRSHOT_TASK_GROUP_2_LABELING_FUNCTION = {
    "operational_outcomes": [
        "guo_los",
        "guo_readmission",
        "guo_icu",
        "mimic4_los",
        "mimic4_mortality",
        "mimic4_readmission",
    ],
    "lab_values": [
        "lab_thrombocytopenia",
        "lab_hyperkalemia",
        "lab_hypoglycemia",
        "lab_hyponatremia",
        "lab_anemia"
    ],
    "new_diagnoses": [
        "new_hypertension",
        "new_hyperlipidemia",
        "new_pancan",
        "new_celiac",
        "new_lupus",
        "new_acutemi"
    ],
    "chexpert": [
        "chexpert"
    ],
}

#############################################
#
# Helper functions
#
#############################################

def copy_file(src: str, dest: str, is_overwrite_if_exists: bool = False) -> None:
    """Copy a file or directory if it does not exist."""
    if is_overwrite_if_exists or not os.path.exists(os.path.join(dest, os.path.basename(src))):
        if os.path.isdir(src):
            logger.info(f"Copying directory from `{src}` to `{dest}`.")
            os.system(f'cp -r {src} {dest}')
        else:
            logger.info(f"Copying file from `{src}` to `{dest}`.")
            os.system(f'cp {src} {dest}')

def copy_resources_to_local(base_dir: str, is_overwrite_if_exists: bool = False) -> None:
    """Copy resources to local-scratch directories."""
    os.makedirs(base_dir, exist_ok=True)
    if base_dir == GPU_BASE_DIR:
        # Don't do any copying if GPU partiton b/c just using the shared drive for now
        return
    # copy_file('/share/pi/nigam/data/som-rit-phi-starr-prod.starr_omop_cdm5_deid_2023_08_13_extract_v9', base_dir, is_overwrite_if_exists=False)
    copy_file('/share/pi/nigam/data/som-rit-phi-starr-prod.starr_omop_cdm5_deid_2023_02_08_extract_v8_no_notes', base_dir, is_overwrite_if_exists=False)
    copy_file('/share/pi/nigam/mwornow/meds_dev', base_dir, is_overwrite_if_exists=False)
    
def rewrite_paths_for_carina_from_config(config: DictConfig) -> DictConfig:
    """Rewrite paths for Carina partitions to use local-scratch directories."""
    if os.environ.get('SLURM_JOB_PARTITION') == 'nigam-v100':
        copy_resources_to_local(V100_BASE_DIR, is_overwrite_if_exists=True)
        if hasattr(config.data.dataset, 'path_to_femr_extract'):
            config.data.dataset.path_to_femr_extract = config.data.dataset.path_to_femr_extract.replace('/share/pi/nigam/data/', V100_BASE_DIR)
        if hasattr(config.data.dataset, 'path_to_meds_reader_extract'):
            config.data.dataset.path_to_meds_reader_extract = config.data.dataset.path_to_meds_reader_extract.replace('/share/pi/nigam/data/', V100_BASE_DIR)
        logger.info(f"Loading data from local-scratch: `{V100_BASE_DIR}`.")
    elif os.environ.get('SLURM_JOB_PARTITION') == 'nigam-a100':
        copy_resources_to_local(A100_BASE_DIR, is_overwrite_if_exists=True)
        if hasattr(config.data.dataset, 'path_to_femr_extract'):
            config.data.dataset.path_to_femr_extract = config.data.dataset.path_to_femr_extract.replace('/share/pi/nigam/data/', A100_BASE_DIR)
        if hasattr(config.data.dataset, 'path_to_meds_reader_extract'):
            config.data.dataset.path_to_meds_reader_extract = config.data.dataset.path_to_meds_reader_extract.replace('/share/pi/nigam/data/', A100_BASE_DIR)
        logger.info(f"Loading data from local-scratch: `{A100_BASE_DIR}`.")
    elif os.environ.get('SLURM_JOB_PARTITION') == 'nigam-h100':
        copy_resources_to_local(H100_BASE_DIR, is_overwrite_if_exists=True)
        if hasattr(config.data.dataset, 'path_to_femr_extract'):
            config.data.dataset.path_to_femr_extract = config.data.dataset.path_to_femr_extract.replace('/share/pi/nigam/data/', H100_BASE_DIR)
        if hasattr(config.data.dataset, 'path_to_meds_reader_extract'):
            config.data.dataset.path_to_meds_reader_extract = config.data.dataset.path_to_meds_reader_extract.replace('/share/pi/nigam/data/', H100_BASE_DIR)
        logger.info(f"Loading data from local-scratch: `{H100_BASE_DIR}`.")
    elif os.environ.get('SLURM_JOB_PARTITION') == 'gpu':
        copy_resources_to_local(GPU_BASE_DIR, is_overwrite_if_exists=True)
        if hasattr(config.data.dataset, 'path_to_femr_extract'):
            config.data.dataset.path_to_femr_extract = config.data.dataset.path_to_femr_extract.replace('/share/pi/nigam/data/', GPU_BASE_DIR)
        if hasattr(config.data.dataset, 'path_to_meds_reader_extract'):
            config.data.dataset.path_to_meds_reader_extract = config.data.dataset.path_to_meds_reader_extract.replace('/share/pi/nigam/data/', GPU_BASE_DIR)
        logger.info(f"Loading data from local-scratch: `{GPU_BASE_DIR}`.")
    elif socket.gethostname() == "bmir-p02.stanford.edu":
        if hasattr(config.data.dataset, 'path_to_femr_extract'):
            config.data.dataset.path_to_femr_extract = config.data.dataset.path_to_femr_extract.replace('/share/pi/nigam/data/', SHAHLAB_SECURE_BASE_DIR)
        if hasattr(config.data.dataset, 'path_to_meds_reader_extract'):
            config.data.dataset.path_to_meds_reader_extract = config.data.dataset.path_to_meds_reader_extract.replace('/share/pi/nigam/data/', SHAHLAB_SECURE_BASE_DIR)
        if hasattr(config.data.tokenizer, 'path_to_config'):
            config.data.tokenizer.path_to_config = config.data.tokenizer.path_to_config.replace('/share/pi/nigam/mwornow/hf_ehr/cache/', SHAHLAB_SECURE_BASE_DIR)
        logger.info(f"Loading data from /home/migufuen: `{SHAHLAB_SECURE_BASE_DIR}`.")
    else:
        logger.info("No local-scratch directory found. Using default `/share/pi/` paths.")
    return config



from setuptools import setup, find_packages

setup(
    name="hf_ehr",
    version="0.1.0",
    author="Michael Wornow",
    packages=find_packages(),
    description="Code for Context Clues paper",
    url="https://github.com/som-shahlab/long_context_clues/",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
)